#! /usr/bin/env swipl -G0M -T0M -L0M -g print_banner -s
% ram is cheap lets use it: set global, trail and local stack to 0 = maximum

:- prolog_load_context(directory,Dir),asserta(user:file_search_path(colp, Dir)). 

:- include(init).

:- op(1100,'xfy',or).	
:- op(1090,'xfy',and).

:- use_module(colp(utils)).
:- use_module(colp(codtree)).
:- use_module(colp(sclause)).
:- use_module(colp(ptree)).
:- use_module(colp(snode)).

examples :-
    nl, 
    print('syntax:'),nl,
    print('?- start(Goal,Programfile).'),nl,
    nl,
    print('examples:'),nl,
    print('?- start(list(cons(_X,cons(_Y,_Z))),\'logic\\listnat.logic\').'),nl,
    nl,
    print('?- start(path(0,P),\'logic\\path.logic\').'),nl,
    nl,nl.

test:-start(list(cons(_X,cons(_Y,_Z))),'logic\\listnat.logic').

start(Goal,FileName):-
    sclause:load_file(FileName,Clauses),
    proof(Clauses,Goal,Result),
    snode:print(Result).

% call without options
proof(Clauses,Goal,Result):-
    proof(Clauses,Goal,[],Result).

% init search queue, set options
proof(Clauses,Goal,Options,Result):-
    ptree:init(Goal,Clauses,PTree),
    snode:init(0,PTree,InitSNode),
    search([InitSNode],Options,Result). 
    
% if success tree found return and continue search for more solutions on backtracking
search([SNode|SNodeList],Options,Result):-
    snode:get_ptree(SNode,PTree),
    ptree:is_success(PTree),!,
    ( Result = SNode
      ;
      search(SNodeList,Options,Result) ).

% the usual breath first search: 
search([SNode|SNodeList],Options,Result):-                    % pick first node in que
    snode:get_ptree(SNode,PTree),                             % unpack from from search node information
    ptree:expand(PTree,Options,PTrees_gen),                   % find residuals and unroll trees
    snode:rescore(PTrees_gen,SNodeList_gen),                  % calculate score to enter into search
    snode:merge_list(SNodeList_gen,SNodeList,SNodesList_new), % merge new trees into queue by score
    search(SNodesList_new,Options,Result).                    % keep going with search
