:- module(codtree,[
                  ]).
%codtree: codtree(or(Head,[and([or(...),...]),...]))

% full open
% or(Head,[],[Clauses]).

% full unrolled
% or(Head,[Clauses],[]).

% dead node
% or(Head,[],[]).


init(Goal,Clauses,CodTree):-
    init_ornode(Goal,OrNode),
    unroll(OrNode,Clauses,CodTree).

init_ornode(Head,or(Head,[],_)).

is_success(CTree):-
    is_success_ornode(CTree),!.
    
is_success_ornode(or(_,AndNodeList,_)):-
    member(and(OrNodeList),AndNodeList),
    all(is_success_ornode,OrNodeList).

unroll(CodTree,Clauses,UnrolledCodTree):-
    unroll_or(CodTree,Clauses,UnrolledCodTree).
 
unroll_or(or(Goal,[],Var),Clauses,OrNode_new):-
    var(Var),!,
    unroll_or(or(Goal,[],Clauses),Clauses,OrNode_new).
    
unroll_or(or(Goal,AndNodeList,Candidates),Clauses,or(Goal,AndNodeList_new,UnifyClauses)):-
    sclause:matches(Candidates,Goal,Matches,UnifyClauses),
    append(AndNodeList,Matches,AndNodeList_tmp),
    unroll_ands(AndNodeList_tmp,Clauses,AndNodeList_new).

unroll_ors([],_Clauses,[]).

unroll_ors([OrNode|OrNodeList],Clauses,[OrNode_new|OrNodeList_new]):-
    unroll_or(OrNode,Clauses,OrNode_new),
    unroll_ors(OrNodeList,Clauses,OrNodeList_new).

unroll_ands([],_Clauses,[]).

unroll_ands([and(OrNodeList)|AndNodeList],Clauses,[and(OrNodeList_new)|AndNodeList_new]):-
    unroll_ors(OrNodeList,Clauses,OrNodeList_new),
    unroll_ands(AndNodeList,Clauses,AndNodeList_new).

simplify_tree(CTree,SimpleTree):-
    simplify_or(CTree,SimpleTree).

simplify_or(or(Goal,AndNodeList,_),Simple):-
    ( simplify_ands(AndNodeList,SimpleAndList) -> Simple = (Goal->{SimpleAndList})
                                                ; Simple = Goal ).

simplify_ors([OrNode],SimpleOrNode):-!,
    simplify_or(OrNode,SimpleOrNode).
                          
simplify_ors([OrNode|OrNodeList],SimpleOrNode and SimpleOrList):-
    simplify_or(OrNode,SimpleOrNode),
    simplify_ors(OrNodeList,SimpleOrList).

simplify_ands([and(OrNodeList)],SimpleOrList):-!,
    simplify_ors(OrNodeList,SimpleOrList).

simplify_ands([and(OrNodeList)|AndNodeList],SimpleOrList or SimpleAndList):-
    simplify_ors(OrNodeList,SimpleOrList),
    simplify_ands(AndNodeList,SimpleAndList).

gather_subst(CodTree,SubstListSorted):-
    find_residual(CodTree,(Goal,Clauses)),
    extract_subst(Clauses,Goal,SubstList),
    sort(SubstList,SubstListSorted).
    %codtree:simplify_tree(CodTree,SimpleTree),
    %print(SimpleTree),nl,
    %print((Goal,Clauses)),nl,
    %print(SubstListSorted),nl.

extract_subst([],_Goal,[]).

extract_subst([Clause|Clauses],Goal,[Substs|SubstList]):-
    sclause:get_head(Clause,Head),
    subst(Goal,Head,Substs),
    extract_subst(Clauses,Goal,SubstList).

subst(Goal,Head,SortedSubsts):-
    copy_term(Head,Head_copy),
    unifiable(Goal,Head_copy,Unifier),
    term_variables(Goal,Vars),
    filter_unifier(Unifier,Vars,Substs),
    sort(Substs,SortedSubsts).

filter_unifier([],_Vars,[]):-!.
    
filter_unifier([X=Y|Tail],Vars,[X=Y|Substs]):-
    smember(X,Vars),!,
    filter_unifier(Tail,Vars,Substs).    

filter_unifier([_|Tail],Vars,Substs):-
    filter_unifier(Tail,Vars,Substs).    

% rightmost on the lowest level in the tree
find_residual(CodTree,Residual):-
    search_residual([CodTree],Residual).

search_residual([or(_Goal,AndNodeList,[])|Queue],Residual):-!,
    gather_ornodes(AndNodeList,Queue,Queue_new),
    search_residual(Queue_new,Residual).

search_residual([or(Goal,_AndNodeList,Clauses)|_Queue],(Head,Clauses)):-
    unify_with_occurs_check(Goal,Head).

gather_ornodes([],OrNodes,OrNodes):-!.

gather_ornodes([and(OrNodeList)|AndNodeList],OrNodes,OrNodes_new):-
    append(OrNodes,OrNodeList,OrNodes_tmp),
    gather_ornodes(AndNodeList,OrNodes_tmp,OrNodes_new).