:- module(ptree,[
                ]).

% ptree contains a coalgebraic derivation tree and some more information:
%
% codtree:       the coalgebraic derivation tree
% clauses:       the clauses that can be used to expand or leafes
% substitutions: the history of substitutions made to the the original goal
% goal:          the original goal to be proven
% root:          the root node of the tree in compact form (term , to be printed)
%
% Substitutions applied to the Goal result in a structural equal term to the Root of the CodTree

%ptree: ptree(Goal,Substs,Root,CodTree,Clauses)
     
init(Goal,Clauses,PTree):-
    codtree:init(Goal,Clauses,CodTree),
    create(Goal,[],Goal,CodTree,Clauses,PTree).

create(Goal,Substs,Root,CodTree,Clauses,ptree(Goal,Substs,Root,CodTree,Clauses)).

get_goal(PTree,Goal):-
    arg(1,PTree,Goal).
    
get_subst(PTree,Subst):-
    arg(2,PTree,Subst).

get_root(PTree,Root):-
    arg(3,PTree,Root).

get_codtree(PTree,CTree):-
    arg(4,PTree,CTree).

get_clauses(PTree,Clauses):-
    arg(5,PTree,Clauses).

is_success(PTree):-
    get_codtree(PTree,CTree),
    codtree:is_success(CTree).

expand(PTree,_Options,PTrees_gen):-
    get_codtree(PTree,CodTree),
    codtree:gather_subst(CodTree,SubstList),
    generate(PTree,SubstList,PTrees_gen).

generate(_PTree,[],[]).

generate(PTree,[AddSubsts|SubstList],[PTree_gen|PTrees_gen]):-
    get_clauses(PTree,Clauses),
    get_codtree(PTree,CodTree),
    get_goal(PTree,Goal),
    get_root(PTree,Root),
    get_subst(PTree,Substs),
    copy_term((Root,CodTree,AddSubsts),(Root_copy,CodTree_copy,AddSubsts_copy)),
    append(Substs,AddSubsts,Substs_new),
    applysubst(AddSubsts_copy),
    codtree:unroll(CodTree_copy,Clauses,UnrolledCodTree),
    create(Goal,Substs_new,Root_copy,UnrolledCodTree,Clauses,PTree_gen),
    generate(PTree,SubstList,PTrees_gen).

applysubst([]).
applysubst([X=Y|Subst]):-
    X=Y,
    applysubst(Subst).