:- module(sclause,[
                  ]).

% module to extract clauses from file and check sanity

load_file(File,SClauses):-
    open(File,read,Stream),
	call_cleanup(read_clauses(Stream,SClauses),
                 close(Stream)).
                         
read_clauses(Stream,[]) :-
        at_end_of_stream(Stream),!.

read_clauses(Stream,SClauses_result) :-
        read_term(Stream,Term,[]),
        (term_to_sclause(Term,SClause) -> 
            SClauses_result = [SClause|SClauses_new]
            ;
            SClauses_result =          SClauses_new
        ),
        read_clauses(Stream,SClauses_new).

% dont allow variables as facts
term_to_sclause(Var,_):-
    var(Var),!,type_error(head,Var).
    
term_to_sclause(Head :- Body,SClause):-
    nonvar(Head),!,
    body_to_orlist(Body,LBody),
    init(Head,LBody,SClause).

term_to_sclause(Head,SClause):-
    nonvar(Head),
    init(Head,[],SClause).

%(A,(B,C)) -> [A,B,C]
body_to_orlist(Var,_):-
    var(Var),!,type_error(bodygoal,Var).

body_to_orlist((Goal,Rest),[OrNode|SRest]):-!,
    codtree:init_ornode(Goal,OrNode),
    body_to_orlist(Rest,SRest).

body_to_orlist(Goal,[OrNode]):-
    codtree:init_ornode(Goal,OrNode).

init(Head,LBody,sclause(Head,and(LBody))).
    

get_head(sclause(Head,_),Head).
get_body(sclause(_,Body),Body).
    
matches([],_Goal,[],[]).

matches([SClause|SClauses],Goal,[Match|Matches],UnifySClauses):-
    get_head(SClause,Head),
    get_body(SClause,Body),
    subsumes_term(Head,Goal),!,
    copy_term((Head,Body),(Goal,Match)), % Instantiate Body to Match, Goal will be not changed because Head subsumes Goal!
    matches(SClauses,Goal,Matches,UnifySClauses).       
    
matches([SClause|SClauses],Goal,Matches,[SClause|UnifySClauses]):-
    get_head(SClause,Head),
    not(not(unify_with_occurs_check(Goal,Head))),!,% check if unifiable without creating cycles, dont do bindings
    matches(SClauses,Goal,Matches,UnifySClauses).

matches([_SClause|SClauses],Goal,Matches,UnifySClauses):-
    matches(SClauses,Goal,Matches,UnifySClauses).